using System;
using System.Drawing;
using System.Collections;
using System.IO;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace _3dedit {
	public struct PointOnFace{
		public double depth;
		public int nface;
		public double u,v;
		public Vector3 WorldPoint;
	}
	public class MeshForm{
		public int NV,NF,NV0,NF0;
		public MYVERTEX []Vert;
		public ushort []Faces;
		public string Name;

		public Vector3 ObjCenter;
		public int FormID;
		public int Count;

		Mesh m_mesh=null;
		
		public MeshForm(MeshForm a){
			NV0=NV=a.NV; NF0=NF=a.NF;
			Vert=new MYVERTEX[NV0];
			for(int i=0;i<NV;i++) Vert[i]=a.Vert[i];
			Faces=new ushort[3*NF0];
			for(int i=0;i<3*NF0;i++) Faces[i]=a.Faces[i];
			Name=a.Name;
			SetCenter();
			Count=0;
		}
		public MeshForm(float [,]pts,ushort[,]faces,string name){
			Name=name;
			int nv=pts.Length/3,nf=faces.Length/3;
			NV0=NV=nv;
			NF0=NF=nf;
			Vert=new MYVERTEX[NV0];
			Faces=new ushort[NF0*3];
			for(int i=0;i<nv;i++){
				Vert[i].p.X=pts[i,0];
				Vert[i].p.Y=pts[i,1];
				Vert[i].p.Z=pts[i,2];
				Vert[i].dc=i==0 ? unchecked((int)0xffc0c040) : unchecked((int)0xffff0000);
			}
			for(int i=0;i<nf;i++){
				Faces[3*i]=faces[i,0];
				Faces[3*i+1]=faces[i,1];
				Faces[3*i+2]=faces[i,2];
			}
			SetNormals();
			m_mesh=null;
			SetCenter();
			Count=0;
		}
		MeshForm(int nv,MYVERTEX []v,int nf,ushort []f){
			NV=NV0=nv;
			NF=NF0=nf;
			Vert=v;
			Faces=f;
			Name=null;
			Count=0;
		}

		void SetNormals(){
			for(int i=0;i<NV;i++){
				Vert[i].n=new Vector3(0,0,0);
			}
			for(int i=0;i<NF;i++){
				int N1=Faces[3*i],N2=Faces[3*i+1],N3=Faces[3*i+2];
				float x=Vert[N1].p.X,y=Vert[N1].p.Y,z=Vert[N1].p.Z;
				float x2=Vert[N2].p.X-x,y2=Vert[N2].p.Y-y,z2=Vert[N2].p.Z-z;
				float x3=Vert[N3].p.X-x,y3=Vert[N3].p.Y-y,z3=Vert[N3].p.Z-z;

				float C0=-y3*z2+y2*z3;
				float C1=-z3*x2+z2*x3;
				float C2=-x3*y2+x2*y3;
				Vert[N1].n.X+=C0; Vert[N1].n.Y+=C1; Vert[N1].n.Z+=C2;
				Vert[N2].n.X+=C0; Vert[N2].n.Y+=C1; Vert[N2].n.Z+=C2;
				Vert[N3].n.X+=C0; Vert[N3].n.Y+=C1; Vert[N3].n.Z+=C2;
			}
			for(int i=0;i<NV;i++){
				Vert[i].n.Normalize();
			}
		}
		public Mesh GetMesh(Device dev){
			if(m_mesh==null){
				m_mesh=(Mesh)S3DirectX.Create3DMesh(dev,NV,Vert,NF*3,Faces);
			}
			return m_mesh;
		}
		public void UpdateMesh(){
			if(m_mesh!=null) m_mesh.Dispose();
			m_mesh=null;
			SetNormals();
		}
		public bool GetBoundBox(ref Matrix tr,out RPoint min, out RPoint max) {
			min = new RPoint(float.MaxValue,float.MaxValue,float.MaxValue);
			max = new RPoint(float.MinValue,float.MinValue,float.MinValue);
			if(NV==0) return false;
			for(int i=0;i<NV;i++){
				Vector3 v=Vert[i].p;
				v.TransformCoordinate(tr);
				RPoint R=new RPoint(v.X,v.Y,v.Z);
				min.SetMin(ref R);
				max.SetMax(ref R);
			}
			return true;
		}
		void SetCenter(){
			Matrix m=Matrix.Identity;
			RPoint ptmin,ptmax;
			GetBoundBox(ref m,out ptmin,out ptmax);
			ObjCenter=new Vector3((float)((ptmin.X+ptmax.X)/2),(float)((ptmin.Y+ptmax.Y)/2),(float)((ptmin.Z+ptmax.Z)/2));
		}
		public float MinLen(int nf){
			int p0=Faces[3*nf],p1=Faces[3*nf+1],p2=Faces[3*nf+2];
			float l0=Vector3.Subtract(Vert[p0].p,Vert[p1].p).Length();
			float l1=Vector3.Subtract(Vert[p1].p,Vert[p2].p).Length();
			float l2=Vector3.Subtract(Vert[p2].p,Vert[p0].p).Length();
			return Math.Min(l0,Math.Min(l1,l2));
		}

		static float[,]_CubeVert={{0,0,0},{0,0,100},{0,100,0},{0,100,100}, {
																			   100,0,0},{100,0,100},{100,100,0},{100,100,100}};
		static ushort[,]_CubeFaces={{0,1,2},{2,1,3},{1,0,4},{1,4,5}, {
																		 0,2,6},{0,6,4},{2,3,7},{2,7,6}, {
																											 3,1,5},{3,5,7},{7,5,4},{7,4,6}};
		public static MeshForm NewMesh(){
			return new MeshForm(_CubeVert,_CubeFaces,null);
		}


		public double GetIntersection(ref Vector3 pt,ref Vector3 dir,out PointOnFace pos,EPickMode pickMode){
			double d0=-1;
			pos=new PointOnFace();
			for(int i=0;i<NF;i++){
				int a0=Faces[3*i],a1=Faces[3*i+1],a2=Faces[3*i+2];
				Matrix M=new Matrix();
				M.M11=Vert[a1].p.X-Vert[a0].p.X;
				M.M21=Vert[a1].p.Y-Vert[a0].p.Y;
				M.M31=Vert[a1].p.Z-Vert[a0].p.Z;
				M.M41=0;
				M.M12=Vert[a2].p.X-Vert[a0].p.X;
				M.M22=Vert[a2].p.Y-Vert[a0].p.Y;
				M.M32=Vert[a2].p.Z-Vert[a0].p.Z;
				M.M42=0;
				M.M13=-dir.X;
				M.M23=-dir.Y;
				M.M33=-dir.Z;
				M.M43=0;
				M.M14=pt.X-Vert[a0].p.X;
				M.M24=pt.Y-Vert[a0].p.Y;
				M.M34=pt.Z-Vert[a0].p.Z;
				M.M44=1;
				float det=1.0f;
				M=Matrix.Invert(ref det,M);
				if(det==0){
					continue;
				}
				double d=-M.M34;
				if(d<0 || (d0>0 && d>d0)) continue;
				double a=-M.M14,b=-M.M24;
				if(a<0 || b<0 || a+b>1) continue;
				bool qok=true;
				switch(pickMode){
					case EPickMode.PickEdge:
						qok=(a<0.02 || b<0.02 || a+b>0.98);
						break;
					case EPickMode.PickPoint:
						qok=(a>0.98 || b>0.98 || a+b<0.02);
						break;
				}
				if(!qok) continue;

				d0=d;
				pos.nface=i; pos.u=a; pos.v=b;
				pos.depth=d;
				pos.WorldPoint.X=(float)(Vert[a0].p.X*(1-a-b)+Vert[a1].p.X*a+Vert[a2].p.X*b);
				pos.WorldPoint.Y=(float)(Vert[a0].p.Y*(1-a-b)+Vert[a1].p.Y*a+Vert[a2].p.Y*b);
				pos.WorldPoint.Z=(float)(Vert[a0].p.Z*(1-a-b)+Vert[a1].p.Z*a+Vert[a2].p.Z*b);
			}
			return d0;
		}

		public bool OnLine(int v1,int v2,int v3){
			Vector3 c1=Vector3.Subtract(Vert[v2].p,Vert[v1].p);
			Vector3 c2=Vector3.Subtract(Vert[v3].p,Vert[v1].p);
			double s0=c1.Y*c2.Z-c1.Z*c2.Y;
			double s1=c1.Z*c2.X-c1.X*c2.Z;
			double s2=c1.X*c2.Y-c1.Y*c2.X;
			double lm=Math.Max(Math.Max(c1.LengthSq(),c2.LengthSq()),Vector3.Subtract(c1,c2).LengthSq());
			return (s0*s0+s1*s1+s2*s2<lm*lm*1e-5);
		}
		public bool HaveEdge(int a,int b){
			for(int i=0;i<NF;i++){
				if((Faces[3*i]==a || Faces[3*i+1]==a || Faces[3*i+2]==a) && 
					(Faces[3*i]==b || Faces[3*i+1]==b || Faces[3*i+2]==b)) return true;
			}
			return false;
		}
		/************* edit form ***********/
		public bool HaveFreeSpace(int nv,int nf){
			return NV<ushort.MaxValue-nv && NF<ushort.MaxValue-nf;
		}
		public int NewVert(Vector3 pt){
			if(NV0==NV){
				NV0=Math.Min(NV+NV/2+1,ushort.MaxValue);
				MYVERTEX []vv=new MYVERTEX[NV0];
				for(int i=0;i<NV;i++) vv[i]=Vert[i];
				Vert=vv;
			}
			if(NV0==NV) return NV-1;
			Vert[NV].p=pt;
			Vert[NV].dc=unchecked((int)0xff808080);
			return NV++;
		}
		public int NewFace(int a,int b,int c){
			if(NF0==NF){
				NF0=Math.Min(NF+NF/2+1,ushort.MaxValue);
				ushort []vv=new ushort[3*NF0];
				for(int i=0;i<3*NF;i++) vv[i]=Faces[i];
				Faces=vv;
			}
			if(NF0==NF) return NF-1;
			Faces[3*NF]=(ushort)a;
			Faces[3*NF+1]=(ushort)b;
			Faces[3*NF+2]=(ushort)c;
			return NF++;
		}
		public void GlueTo(int v0,int v1){
			for(int i=0;i<NF;i++){
				bool c0=false,c1=false;
				for(int j=0;j<3;j++){
					if(Faces[3*i+j]==v1) c1=true;
					else if(Faces[3*i+j]==v0){
						Faces[3*i+j]=(ushort)v1; c0=true;
					}
				}
				if(c0&&c1){
					RemoveFace(i);
					i--;
				}
			}
			NV--;
			if(v0==NV) return;
			for(int i=0;i<3*NF;i++){
				if(Faces[i]==NV) Faces[i]=(ushort)v0;
			}
			Vert[v0]=Vert[NV];
		}
		public void RemoveFace(int i){
			NF--;
			Faces[3*i]=Faces[3*NF];
			Faces[3*i+1]=Faces[3*NF+1];
			Faces[3*i+2]=Faces[3*NF+2];
		}
		public bool GetConjFace(int nf,int np,out int nf1,out int np1){
			nf1=np1=0;
			int a0=Faces[3*nf+np],a1=Faces[3*nf+(np+1)%3];
			for(int i=0;i<NF;i++){
				if(Faces[3*i]==a1 && Faces[3*i+1]==a0){ nf1=i; np1=0; return true; }
				if(Faces[3*i+1]==a1 && Faces[3*i+2]==a0){ nf1=i; np1=1; return true; }
				if(Faces[3*i+2]==a1 && Faces[3*i]==a0){ nf1=i; np1=2; return true; }
			}
			return false;
		}

		public void AlignEdge(int v0,int v1,bool force){
			Vector3 dv=Vector3.Subtract(Vert[v1].p,Vert[v0].p);
			if(!force){
				double a=dv.Length()/12;
				if(Math.Abs(dv.X)<a) dv.X=0;
				if(Math.Abs(dv.Y)<a) dv.Y=0;
				if(Math.Abs(dv.Z)<a) dv.Z=0;
			}else{
				double x=Math.Abs(dv.X),y=Math.Abs(dv.Y),z=Math.Abs(dv.Z);
				int k= (x>y && x>z) ? 0 : y>z ? 1 : 2;
				if(k!=0) dv.X=0;
				if(k!=1) dv.Y=0;
				if(k!=2) dv.Z=0;
			}
			Vert[v1].p=Vector3.Add(Vert[v0].p,dv);

			UpdateMesh();
		}
		public void AlignToLine(int l0,int l1,int v){
			Vector3 dl=Vector3.Subtract(Vert[l1].p,Vert[l0].p);
			Vector3 dp=Vector3.Subtract(Vert[v].p,Vert[l0].p);
			float ssp=(dl.X*dp.X+dl.Y*dp.Y+dl.Z*dp.Z)/dl.LengthSq();
			dl.Scale(ssp);
			Vert[v].p=Vector3.Add(Vert[l0].p,dl);
		}
		public void AlignToPlane(int l0,int l1,int l2,int v){
			Vector3 dl1=Vector3.Subtract(Vert[l1].p,Vert[l0].p);
			Vector3 dl2=Vector3.Subtract(Vert[l2].p,Vert[l0].p);
			Vector3 dp=Vector3.Subtract(Vert[v].p,Vert[l0].p);
			float s0=dl1.Y*dl2.Z-dl1.Z*dl2.Y;
			float s1=dl1.Z*dl2.X-dl1.X*dl2.Z;
			float s2=dl1.X*dl2.Y-dl1.Y*dl2.X;
			Vector3 nl=new Vector3(s0,s1,s2);
			float ssp=(nl.X*dp.X+nl.Y*dp.Y+nl.Z*dp.Z)/nl.LengthSq();
			nl.Scale(ssp);
			Vert[v].p.Subtract(nl);
		}
		/*** IO ***/
		public static MeshForm LoadForm(BinaryReader fp){
			// header
			int NV=fp.ReadInt32();
			int NF=fp.ReadInt32();
			for(int i=0;i<6;i++) fp.ReadInt32();  // reserved

			MYVERTEX []Pts=new MYVERTEX[NV];
			for(int i=0;i<NV;i++){
				Pts[i].p.X=(float)(fp.ReadInt32()/1000.0);
				Pts[i].p.Y=(float)(fp.ReadInt32()/1000.0);
				Pts[i].p.Z=(float)(fp.ReadInt32()/1000.0);
				Pts[i].dc=fp.ReadInt32();
			}
			ushort []Faces=new ushort[3*NF];
			for(int i=0;i<3*NF;i++){
				Faces[i]=fp.ReadUInt16();
			}
			return new MeshForm(NV,Pts,NF,Faces);
		}
		public void SaveForm(BinaryWriter bw){
			bw.Write((Int32)NV);
			bw.Write((Int32)NF);
			for(int i=0;i<6;i++) bw.Write((Int32)0);
			for(int i=0;i<NV;i++){
				bw.Write((Int32)(Vert[i].p.X*1000));
				bw.Write((Int32)(Vert[i].p.Y*1000));
				bw.Write((Int32)(Vert[i].p.Z*1000));
				bw.Write((Int32)(Vert[i].dc));
			}
			for(int i=0;i<3*NF;i++){
				bw.Write((UInt16)Faces[i]);
			}
		}
	}
	
	public class MeshObj : IDXObject {
		MeshForm m_form;
		Matrix m_wm;
		Color m_col;
		bool m_IsSelected=false;
		string m_name;
		bool []m_selectedPoints;
		
		static Mesh m_Sphere;
		static float m_SpRad=8;
		static Vector3 m_selctr;
		bool qSelCtrChanged;
		public string Name{
			get{ return m_form.Name; }
			set{ m_form.Name=value; }
		}
		public MeshForm Form{
			get{ return m_form; }
			set{ m_form=value; }
		}
		public Color BaseColor{
			get{ return m_col; }
			set{ m_col=value; }
		}

		public MeshObj(MeshForm mf,string name) {
			m_wm=Matrix.Identity;
			m_form=mf;
			m_name=name;
			m_col=Color.Gray;
		}

		public MeshObj(MeshObj a){
			m_wm=a.m_wm;
			m_form=new MeshForm(a.m_form);
			m_name=a.m_name;
			m_col=a.m_col;
		}

		public static MeshObj NewMesh(){
			return new MeshObj(MeshForm.NewMesh(),null);
		}
	
		public bool IsSelected{
			get{ return m_IsSelected; }
			set{ m_IsSelected=value; }
		}
		public void RotateObject(ref Vector3 pt, ref Vector3 vec,double ang){ // rotate, slide, click, moveto (for camera) 
			if(ang==0) return;
			Vector3 ctr=m_form.ObjCenter;
			ctr.TransformCoordinate(m_wm);
			MkConjShift(ref ctr,1.0f);
			MkRotate(ref vec,-(float)ang);
			MkConjShift(ref ctr,-1.0f);
		}
		public void MoveObject(ref Vector3 vec){ // rotate, slide, click, moveto (for camera) 
			MkShift(ref vec);
		}
		public void ScaleObject(double cf){ // rotate, slide, click, moveto (for camera) 
			if(cf==1) return;
			Vector3 ctr=m_form.ObjCenter;
			for(int i=0;i<m_form.NV;i++){
				Vector3 v=m_form.Vert[i].p;
				v.Subtract(ctr);
				v.Scale((float)cf);
				v.Add(ctr);
				m_form.Vert[i].p=v;
			}
			m_form.UpdateMesh();
		}
		public void Render(S3DirectX d3dDevice) {
			d3dDevice.SetupObjectMaterial(m_IsSelected ? Color.White : m_col);
			d3dDevice.Renderer.SetTransform(TransformType.World, m_wm);
			Mesh bm=m_form.GetMesh(d3dDevice.DxDevice);
			bm.DrawSubset(0);
			if(m_selectedPoints!=null){
				if(m_Sphere==null){
					m_Sphere=Mesh.Sphere(d3dDevice.DxDevice,m_SpRad,8,4);
				}
				d3dDevice.SetupObjectMaterial(Color.LightPink);
				int ls=Math.Min(m_form.NV,m_selectedPoints.Length);
				for(int a=0;a<ls;a++){
					if(m_selectedPoints[a]){
						Vector3 v=m_form.Vert[a].p;
						v.TransformCoordinate(m_wm);
						Matrix m=Matrix.Translation(v);
						d3dDevice.Renderer.SetTransform(TransformType.World, m);
						m_Sphere.DrawSubset(0);
					}
				}
			}
			d3dDevice.Renderer.SetTransform(TransformType.World,Matrix.Identity);
		}

		public bool GetBounds(out RPoint min, out RPoint max) {
			return m_form.GetBoundBox(ref m_wm,out min,out max);
		}

		public void MkShift(ref Vector3 Shift) {
			Matrix trans = Matrix.Translation(Shift.X,Shift.Y,Shift.Z);
			ApplyMatrix(ref trans);
		}
		public void ApplyMatrix(ref Matrix mat) {
			m_wm = Matrix.Multiply(m_wm, mat);
		}
		public void MkConjShift(ref Vector3 Shift,float cf) {
			Matrix trans = Matrix.Translation(-Shift.X*cf,-Shift.Y*cf,-Shift.Z*cf);
			ApplyMatrix(ref trans);
		}

		public void MkRotate(ref Vector3 vec,float ang) {
			if(ang==0) return;
			Matrix rot;
			rot = Matrix.RotationAxis(vec,ang);
			ApplyMatrix(ref rot);
		}

		public void CleanUp() {
			m_form=null;
		}

		public bool NeedLightning {
			get {
				return true;
			}
		}

		public static void ScaleSphere(double cf){
			m_SpRad*=(float)cf;
			m_Sphere=null;
		}



		public double GetIntersection(ref Vector3 pt,ref Vector3 dir,out PointOnFace pos,EPickMode pickMode){
			pos=new PointOnFace();
			Matrix tr1=Matrix.Invert(m_wm);
			Vector3 pt1=pt; pt1.Add(dir);
			Vector3 pt0=pt;
			pt0.TransformCoordinate(tr1);
			pt1.TransformCoordinate(tr1);
			pt1.Subtract(pt0);
			double d=m_form.GetIntersection(ref pt0,ref pt1,out pos,pickMode);
			if(d>=0) pos.WorldPoint.TransformCoordinate(m_wm);
			return d;
		}

		public void UnselectAll(){
			m_selectedPoints=null;
			m_Sphere=null;
		}
		public void RotateSelectedPoints(ref Vector3 pt, ref Vector3 vec,double ang){ // rotate, slide, click, moveto (for camera) 
			if(m_selectedPoints==null || ang==0) return;
			Matrix m1=m_wm;
			m1.M14=m1.M24=m1.M34=0;
			m1.Invert();
			Vector3 v=Vector3.TransformCoordinate(vec,m1);
			Matrix rot=Matrix.RotationAxis(vec,-(float)ang);
			SetSelCtr();
			int ls=Math.Min(m_form.NV,m_selectedPoints.Length);
			for(int i=0;i<ls;i++){
				if(m_selectedPoints[i]){
					Vector3 p=m_form.Vert[i].p;
					p.Subtract(m_selctr);
					p.TransformCoordinate(rot);
					p.Add(m_selctr);
					m_form.Vert[i].p=p;
				}
			}
			m_form.UpdateMesh();
		}
		public void MoveSelectedPoints(ref Vector3 vec){ // rotate, slide, click, moveto (for camera) 
			if(m_selectedPoints==null) return;
			Matrix m1=m_wm;
			m1.M41=m1.M42=m1.M43=0;
			m1.Invert();
			Vector3 v=Vector3.TransformCoordinate(vec,m1);
			int ls=Math.Min(m_form.NV,m_selectedPoints.Length);
			for(int i=0;i<ls;i++){
				if(m_selectedPoints[i]){
					m_form.Vert[i].p.Add(v);
				}
			}
			m_form.UpdateMesh();
		}
		public int NSelectedPoints(){
			if(m_selectedPoints==null) return 0;
			int ls=Math.Min(m_form.NV,m_selectedPoints.Length);
			int c=0;
			for(int i=0;i<ls;i++) if(m_selectedPoints[i]) c++;
			return c;
		}
		public void ScaleSelectedPoints(double cf){ // rotate, slide, click, moveto (for camera) 
			if(NSelectedPoints()==0 || cf==1) return;
			SetSelCtr();

			int ls=Math.Min(m_form.NV,m_selectedPoints.Length);
			for(int i=0;i<ls;i++){
				if(m_selectedPoints[i]){
					Vector3 v=m_form.Vert[i].p;
					v.Subtract(m_selctr);
					v.Scale((float)cf);
					v.Add(m_selctr);
					m_form.Vert[i].p=v;
				}
			}
			m_form.UpdateMesh();
		}
		void SetSelectedLength(){
			int np=m_form.NV0;
			if(m_selectedPoints==null){
				m_selectedPoints=new bool[np];
				return;
			}
			int L=m_selectedPoints.Length;
			if(L<m_form.NV){
				bool []t=new bool[np];
				for(int i=0;i<L;i++) t[i]=m_selectedPoints[i];
				m_selectedPoints=t;
			}
		}
		public void FillSelected(int nf,int np){
			qSelCtrChanged=true;
			SetSelectedLength();
			int pt=m_form.Faces[3*nf+np];
			if(m_selectedPoints[pt]) return;
			ArrayList newpt=new ArrayList();
			ArrayList oldpt=new ArrayList();
			oldpt.Add(pt);
			m_selectedPoints[pt]=true;
			while(oldpt.Count!=0){
				for(int i=0;i<3*m_form.NF;i++){
					int a=m_form.Faces[i];
					if(oldpt.Contains(a) || newpt.Contains(a)){
						int m=i/3;
						for(int j=0;j<3;j++){
							if(3*m+j!=i){
								int b=m_form.Faces[3*m+j];
								if(!m_selectedPoints[b]){
									newpt.Add(b);
									m_selectedPoints[b]=true;
								}
							}
						}
					}
				}
				oldpt=newpt; newpt=new ArrayList();
			}
		}
		ArrayList GetSelectedFaces(){
			ArrayList res=new ArrayList();
			ushort []ff=m_form.Faces;
			for(int i=0;i<m_form.NF;i++){
				if(m_selectedPoints[ff[3*i]] &&
					m_selectedPoints[ff[3*i+1]] &&
					m_selectedPoints[ff[3*i+2]]){
					res.Add(i);
				}
			}
			return res;

		}
		public void MakeHole(){
			if(NSelectedPoints()!=6) return;
			if(!m_form.HaveFreeSpace(0,4)) return;
			ArrayList sf=GetSelectedFaces();
			if(sf.Count!=2) return;
			ushort []ff=m_form.Faces;
			int f1=(int)sf[0],f2=(int)sf[1];
			double dmax=double.MaxValue;
			int i1=0,j1=0;
			Vector3 ss=m_form.Vert[ff[3*f1]].p;
			ss.Add(m_form.Vert[ff[3*f1+1]].p);
			ss.Add(m_form.Vert[ff[3*f1+2]].p);
			ss.Subtract(m_form.Vert[ff[3*f2]].p);
			ss.Subtract(m_form.Vert[ff[3*f2+1]].p);
			ss.Subtract(m_form.Vert[ff[3*f2+2]].p);
			for(int i=0;i<3;i++){
				int a=ff[3*f1+i];
				for(int j=0;j<3;j++){
					int b=ff[3*f2+j];
					Vector3 v=m_form.Vert[a].p;
					v.Subtract(m_form.Vert[b].p);
					double d=(v.X*ss.X+v.Y*ss.Y+v.Z*ss.Z)/v.Length();
					if(d<dmax){ dmax=d; i1=i; j1=j; }
				}
			}
			int a1=ff[3*f1+i1],b1=ff[3*f1+(i1+1)%3],c1=ff[3*f1+(i1+2)%3];
			int a2=ff[3*f2+j1],b2=ff[3*f2+(j1+1)%3],c2=ff[3*f2+(j1+2)%3];
			ff[3*f1+i1]=(ushort)a2; ff[3*f2+j1]=(ushort)a1;
			m_form.NewFace(a1,b1,b2);
			m_form.NewFace(a1,c2,c1);
			m_form.NewFace(a2,c1,c2);
			m_form.NewFace(a2,b2,b1);
			m_form.UpdateMesh();
		}
		bool IsPointSelected(int k){
			return !(m_selectedPoints==null || k>=m_selectedPoints.Length || !m_selectedPoints[k]);
		}			
		int SelectedPointIndex(int k){
			if(!IsPointSelected(k)) return -1;
			int c=0;
			for(int i=0;i<k;i++) if(m_selectedPoints[i]) c++;
			return c;
		}
		int []GetSelectedPoints(int lm){
			int []a=new int[lm];
			int ls=Math.Min(m_form.NV,m_selectedPoints.Length);
			int c=0;
			for(int i=0;i<ls;i++){
				if(m_selectedPoints[i]){
					a[c++]=i;
					if(c==lm) break;
				}
			}
			return a;
		}

		public void CutHole(){
			if(NSelectedPoints()!=6) return;
			ArrayList sf=GetSelectedFaces();
			if(sf.Count!=6) return;
			ushort []ff=m_form.Faces;
			int[,]M=new int[6,6];
			int []spind=new int[6];
			for(int i=0;i<6;i++){
				int f=(int)sf[i];
				int a=SelectedPointIndex((int)ff[3*f]);
				int b=SelectedPointIndex((int)ff[3*f+1]);
				int c=SelectedPointIndex((int)ff[3*f+2]);
				spind[a]=ff[3*f]; spind[b]=ff[3*f+1]; spind[c]=ff[3*f+2]; 
				M[a,b]|=1; M[b,c]|=1; M[c,a]|=1;
				M[a,c]|=2; M[c,b]|=2; M[b,a]|=2;
			}
			int[]M1=new int[6];
			for(int i=0;i<6;i++){
				for(int j=0;j<6;j++){
					if(M[i,j]==2) M1[j]=i;
				}
			}
			int f1=(int)sf[0];
			for(int i=0;i<6;i++){
				if(M1[i]>=0){
					ff[f1*3]=(ushort)spind[i];
					int j=M1[i]; M1[i]=-1;
					ff[f1*3+1]=(ushort)spind[j];
					int k=M1[j]; M1[j]=-1;
					ff[f1*3+2]=(ushort)spind[k];
					M1[k]=-1;
					f1=(int)sf[1];
				}
			}

			for(int i=6;--i>=2;) m_form.RemoveFace((int)sf[i]);
			m_form.UpdateMesh();
		
		}
		public void Smooth(){
			if(NSelectedPoints()==0) return;
			int ls=Math.Min(m_form.NV,m_selectedPoints.Length);
			Vector3[]v=new Vector3[ls];
			int[]nb=new int[ls];
			for(int i=0;i<ls;i++){
				if(m_selectedPoints[i]){
					v[i]=Vector3.Scale(m_form.Vert[i].p,4);
					nb[i]=4;
				}
			}
			for(int i=0;i<m_form.NF;i++){
				int a=m_form.Faces[3*i];
				int b=m_form.Faces[3*i+1];
				int c=m_form.Faces[3*i+2];
				if(a<ls && m_selectedPoints[a]){
					v[a].Add(m_form.Vert[b].p);
					v[a].Add(m_form.Vert[c].p);
					nb[a]+=2;
				}
				if(b<ls && m_selectedPoints[b]){
					v[b].Add(m_form.Vert[a].p);
					v[b].Add(m_form.Vert[c].p);
					nb[b]+=2;
				}
				if(c<ls && m_selectedPoints[c]){
					v[c].Add(m_form.Vert[a].p);
					v[c].Add(m_form.Vert[b].p);
					nb[c]+=2;
				}
			}

			for(int i=0;i<ls;i++){
				if(m_selectedPoints[i]){
					m_form.Vert[i].p=Vector3.Scale(v[i],(float)(1.0/nb[i]));
				}
			}

			m_form.UpdateMesh();
		}
		public void AlignEdge(int nf,int np,double par){
			int a0=m_form.Faces[3*nf+np%3];
			int a1=m_form.Faces[3*nf+(np+1)%3];
			if(par<0.5) m_form.AlignEdge(a1,a0,true);
			else m_form.AlignEdge(a0,a1,true);
		}

		public void AlignToLine(int nf,int np,double par){
			int N=NSelectedPoints();
			int a0=m_form.Faces[3*nf+np%3];
			int a1=m_form.Faces[3*nf+(np+1)%3];
			int a=par<0.5 ? a0 : a1;
			if(N==0){
				if(par<0.5) m_form.AlignEdge(a1,a0,false);
				else m_form.AlignEdge(a0,a1,false);
			}else{
				int []sp=GetSelectedPoints(N);
				switch(N){
					case 1:
						m_form.AlignEdge(sp[0],a,false);
						break;
					case 2:
						m_form.AlignToLine(sp[0],sp[1],a);
						break;
					case 3:
						m_form.AlignToPlane(sp[0],sp[1],sp[2],a);
						break;
				}

			}
			m_form.UpdateMesh();
		}

		public void PaintSelected(Color col){
			if(NSelectedPoints()==0) return;
			int ls=Math.Min(m_form.NV,m_selectedPoints.Length);
			for(int c=0;c<ls;c++){
				if(m_selectedPoints[c]){
					m_form.Vert[c].dc=col.ToArgb();
				}
			}
			m_form.UpdateMesh();
		}

		void SetSelCtr(){			
			if(!qSelCtrChanged || NSelectedPoints()==0) return;
			Vector3 v=new Vector3(0,0,0);
			int N=0;
			int ls=Math.Min(m_form.NV,m_selectedPoints.Length);
			for(int k=0;k<ls;k++){
				if(m_selectedPoints[k]){
					v.Add(m_form.Vert[k].p);
					N++;
				}
			}
			if(N!=0){
				v.Scale((float)(1.0/N));
			}
			m_selctr=v;
			qSelCtrChanged=false;
		}
		public void SelectPoint(int nf,int np){
			qSelCtrChanged=true;
			SetSelectedLength();
			int p0=m_form.Faces[3*nf+np];
			m_selectedPoints[p0]=!m_selectedPoints[p0];
		}
		public void SplitEdge(int nf,int np,double par){
			if(!m_form.HaveFreeSpace(1,2)) return;
			int nf1,np1;
			bool r=m_form.GetConjFace(nf,np,out nf1,out np1);
			if(!r) return;
			int a0=m_form.Faces[3*nf+(np)%3];
			int a1=m_form.Faces[3*nf+(np+1)%3];
			int a2=m_form.Faces[3*nf+(np+2)%3];
			int a3=m_form.Faces[3*nf1+(np1+2)%3];
			// [a0,a1,a2],[a1,a0,a3] => [a0,a4,a2],[a4,a1,a2],[a4,a0,a3],[a1,a4,a3]
			Vector3 v=m_form.Vert[a1].p;
			v.Subtract(m_form.Vert[a0].p);
			v.Scale((float)par);
			v.Add(m_form.Vert[a0].p);
			int a4=m_form.NewVert(v);
			m_form.Faces[3*nf+(np+1)%3]=m_form.Faces[3*nf1+np1%3]=(ushort)a4;
			m_form.NewFace(a4,a1,a2);
			m_form.NewFace(a1,a4,a3);
			SelectPoint(nf1,np1%3);
			m_form.UpdateMesh();
		}
		public void AddPoint(int nf,double u,double v){
			if(!m_form.HaveFreeSpace(1,2)) return;
			int a0=m_form.Faces[3*nf];
			int a1=m_form.Faces[3*nf+1];
			int a2=m_form.Faces[3*nf+2];
			// [a0,a1,a2] => [a0,a1,a3], [a0,a3,a2], [a3,a1,a2]
			Vector3 v1=Vector3.Subtract(m_form.Vert[a1].p,m_form.Vert[a0].p);
			v1.Scale((float)u);
			Vector3 v2=Vector3.Subtract(m_form.Vert[a2].p,m_form.Vert[a0].p);
			v2.Scale((float)v);
			v1.Add(v2);
			v1.Add(m_form.Vert[a0].p);
			int a3=m_form.NewVert(v1);

			m_form.Faces[3*nf+2]=(ushort)a3;
			m_form.NewFace(a3,a1,a2);
			m_form.NewFace(a0,a3,a2);
			SelectPoint(nf,2);
			m_form.UpdateMesh();
		}
		public void FlipEdge(int nf,int np,double par){
			if(!m_form.HaveFreeSpace(1,2)) return;
			int nf1,np1;
			bool r=m_form.GetConjFace(nf,np,out nf1,out np1);
			if(!r) return;
			int a0=m_form.Faces[3*nf+(np)%3];
			int a1=m_form.Faces[3*nf+(np+1)%3];
			int a2=m_form.Faces[3*nf+(np+2)%3];
			int a3=m_form.Faces[3*nf1+(np1+2)%3];
			if(m_form.OnLine(a0,a3,a2) || m_form.OnLine(a1,a3,a2) || m_form.HaveEdge(a2,a3)) return;
			// [a0,a1,a2],[a1,a0,a3] => [a0,a3,a2],[a1,a2,a3]
			m_form.Faces[3*nf+(np+1)%3]=(ushort)a3;
			m_form.Faces[3*nf1+(np1+1)%3]=(ushort)a2;
			m_form.UpdateMesh();
		}
		public void ShrinkEdge(int nf,int np,double par){
			UnselectAll();
			int a0=m_form.Faces[3*nf+np%3];
			int a1=m_form.Faces[3*nf+(np+1)%3];
			if(par<0.5) m_form.GlueTo(a0,a1);
			else m_form.GlueTo(a1,a0);
			m_form.UpdateMesh();
		}

		/********* IO ***********/

		const int FILE_TAG= 0x4048534D;
		public static MeshObj Load(string fn){
			try{
				BinaryReader br=new BinaryReader(new FileStream(fn,FileMode.Open));
				int tag=br.ReadInt32();
				if(tag!=FILE_TAG) return null;
				int col=br.ReadInt32();
				MeshForm mf=MeshForm.LoadForm(br);
				if(mf==null) return null;
				MeshObj mr=new MeshObj(mf,null);
				mr.m_col=Color.FromArgb(col);
				br.Close();
				mr.m_form.UpdateMesh();
				return mr;
			}catch{
				return null;
			}
		}
		public void Save(string fn){
			try{
				BinaryWriter br=new BinaryWriter(new FileStream(fn,FileMode.Create));
				br.Write((Int32)FILE_TAG);
				br.Write((Int32)m_col.ToArgb());
				m_form.SaveForm(br);
				br.Close();
			}catch{
			}

		}

		public static MeshObj LoadInfo(BinaryReader br,MeshForm []mff){
			int formid=br.ReadInt32();			
			MeshForm mf=mff[formid];
			
			MeshObj res=new MeshObj(mf,null);
			int col=br.ReadInt32();
			res.m_col=Color.FromArgb(col);
			Matrix M=new Matrix();
			br.ReadInt32(); br.ReadInt32();
			M.M11=br.ReadSingle(); M.M12=br.ReadSingle(); M.M13=br.ReadSingle();  
			M.M21=br.ReadSingle(); M.M22=br.ReadSingle(); M.M23=br.ReadSingle(); 
			M.M31=br.ReadSingle(); M.M32=br.ReadSingle(); M.M33=br.ReadSingle(); 
			M.M41=br.ReadSingle(); M.M42=br.ReadSingle(); M.M43=br.ReadSingle(); 
			M.M14=M.M24=M.M34=0; M.M44=1;
			res.m_wm=M;
			res.m_form.UpdateMesh();
			return res;
		}
			
		public void SaveInfo(BinaryWriter br){
			br.Write((Int32)(m_form.FormID));
			br.Write((Int32)(m_col.ToArgb()));
			br.Write((Int32)0); br.Write((Int32)0);
			br.Write(m_wm.M11); br.Write(m_wm.M12); br.Write(m_wm.M13); 
			br.Write(m_wm.M21); br.Write(m_wm.M22); br.Write(m_wm.M23);
			br.Write(m_wm.M31); br.Write(m_wm.M32); br.Write(m_wm.M33);
			br.Write(m_wm.M41); br.Write(m_wm.M42); br.Write(m_wm.M43);
		}
	
	}
}
